﻿/*
 * Thread Proxy Mediator Pattern Example
 * Copyright (C) 2008 Michael Birken
 * 
 * This file is part of Thread Proxy Mediator Pattern Example.
 *
 * Thread Proxy Mediator Pattern Example is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published 
 * by the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Thread Proxy Mediator Pattern Example is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Reflection;
using System.Reflection.Emit;

namespace WindowsFormsExample {
  public sealed class ThreadProxyFactory {

    private ThreadProxyFactory() {
    }

    public static T createProxy<T>(T target, Control control) {

      Type interfaceType = typeof(T);
      Type controlType = typeof(Control);
      MethodInfo beginInvokeMethodInfo = controlType.GetMethod("BeginInvoke", new Type[] { typeof(Delegate) });

      AssemblyBuilder assemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(
          new AssemblyName("WorkerThreadProxyAssembly"),
          AssemblyBuilderAccess.Run);
      ModuleBuilder moduleBuilder = assemblyBuilder.DefineDynamicModule(
          "WorkerThreadProxyAssembly.ControlThreadProxyModule");

      TypeBuilder controlDelegateTypeBuilder = moduleBuilder.DefineType(
          "ThreadProxyFactoryNamespace.0ControlDelegate",
          TypeAttributes.Public | TypeAttributes.AutoClass | TypeAttributes.AnsiClass | TypeAttributes.Sealed,
          typeof(MulticastDelegate));
      ConstructorBuilder controlDelegateConstructorBuilder = controlDelegateTypeBuilder.DefineConstructor(
          MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.SpecialName
              | MethodAttributes.RTSpecialName,
          CallingConventions.Standard,
          new Type[] { typeof(object), typeof(IntPtr) });
      controlDelegateConstructorBuilder.SetImplementationFlags(
          MethodImplAttributes.Runtime | MethodImplAttributes.Managed);
      MethodBuilder controlDelegateInvokeMethodBuilder = controlDelegateTypeBuilder.DefineMethod(
          "Invoke",
          MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot
              | MethodAttributes.Virtual,
          typeof(void),
          new Type[0]);
      controlDelegateInvokeMethodBuilder.SetImplementationFlags(
          MethodImplAttributes.Runtime | MethodImplAttributes.Managed);
      MethodBuilder controlDelegateBeginInvokeMethodBuilder = controlDelegateTypeBuilder.DefineMethod(
          "BeginInvoke",
          MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot
              | MethodAttributes.Virtual,
          typeof(IAsyncResult),
          new Type[] { typeof(AsyncCallback), typeof(object) });
      controlDelegateBeginInvokeMethodBuilder.SetImplementationFlags(
          MethodImplAttributes.Runtime | MethodImplAttributes.Managed);
      MethodBuilder controlDelegateEndInvokeMethodBuilder = controlDelegateTypeBuilder.DefineMethod(
          "EndInvoke",
          MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot
              | MethodAttributes.Virtual,
          typeof(void),
          new Type[] { typeof(IAsyncResult) });
      controlDelegateEndInvokeMethodBuilder.SetImplementationFlags(
          MethodImplAttributes.Runtime | MethodImplAttributes.Managed);
      controlDelegateTypeBuilder.CreateType();

      TypeBuilder typeBuilder = moduleBuilder.DefineType(
          "ThreadProxyFactoryNamespace.0ClassProxy",
          TypeAttributes.Class | TypeAttributes.Public,
          typeof(Object),
          new Type[] { interfaceType });

      FieldBuilder targetField = typeBuilder.DefineField("target", interfaceType, FieldAttributes.Private);
      FieldBuilder controlField = typeBuilder.DefineField("control", controlType, FieldAttributes.Private);
      ConstructorBuilder constructorBuilder = typeBuilder.DefineConstructor(
          MethodAttributes.Public | MethodAttributes.HideBySig
              | MethodAttributes.SpecialName | MethodAttributes.RTSpecialName,
          CallingConventions.HasThis | CallingConventions.ExplicitThis,
          new Type[] { interfaceType, controlType });
      ILGenerator ilGenerator = constructorBuilder.GetILGenerator();
      ilGenerator.Emit(OpCodes.Ldarg_0);
      ilGenerator.Emit(OpCodes.Call, typeof(object).GetConstructor(new Type[0]));
      ilGenerator.Emit(OpCodes.Ldarg_0);
      ilGenerator.Emit(OpCodes.Ldarg_1);
      ilGenerator.Emit(OpCodes.Stfld, targetField);
      ilGenerator.Emit(OpCodes.Ldarg_0);
      ilGenerator.Emit(OpCodes.Ldarg_2);
      ilGenerator.Emit(OpCodes.Stfld, controlField);
      ilGenerator.Emit(OpCodes.Ret);

      foreach (MethodInfo methodInfo in interfaceType.GetMethods()) {
        ParameterInfo[] parameterInfos = methodInfo.GetParameters();
        Type[] parameters = new Type[parameterInfos.Length];
        for (int i = 0; i < parameterInfos.Length; i++) {
          parameters[i] = parameterInfos[i].ParameterType;
        }

        string callClassName = "ThreadProxyFactoryNamespace.00Class" + methodInfo.Name;
        TypeBuilder callClassTypeBuilder = moduleBuilder.DefineType(
            callClassName, TypeAttributes.Class | TypeAttributes.Public);
        constructorBuilder = callClassTypeBuilder.DefineDefaultConstructor(MethodAttributes.Public);
        MethodBuilder runMethodBuilder = typeBuilder.DefineMethod(
            "Run",
            MethodAttributes.Public | MethodAttributes.HideBySig,
            CallingConventions.HasThis | CallingConventions.ExplicitThis,
            typeof(void),
            new Type[0]);
        ilGenerator = runMethodBuilder.GetILGenerator();
        ilGenerator.Emit(OpCodes.Ldarg_0);
        FieldBuilder innerTargetField
            = callClassTypeBuilder.DefineField("0target", interfaceType, FieldAttributes.Public);
        ilGenerator.Emit(OpCodes.Ldfld, innerTargetField);
        FieldBuilder[] innerField = new FieldBuilder[parameterInfos.Length];
        for (int i = 0; i < parameterInfos.Length; i++) {
          ParameterInfo parameterInfo = parameterInfos[i];
          innerField[i] = callClassTypeBuilder.DefineField(
              parameterInfo.Name, parameterInfo.ParameterType, FieldAttributes.Public);
          ilGenerator.Emit(OpCodes.Ldarg_0);
          ilGenerator.Emit(OpCodes.Ldfld, innerField[i]);
        }
        ilGenerator.Emit(OpCodes.Callvirt, methodInfo);
        ilGenerator.Emit(OpCodes.Ret);

        Type callClassType = callClassTypeBuilder.CreateType();

        MethodBuilder methodBuilder = typeBuilder.DefineMethod(
            methodInfo.Name,
            MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot
                | MethodAttributes.Virtual | MethodAttributes.Final,
            CallingConventions.HasThis | CallingConventions.ExplicitThis,
            methodInfo.ReturnType,
            parameters);
        ilGenerator = methodBuilder.GetILGenerator();
        ilGenerator.DeclareLocal(callClassType);
        ilGenerator.Emit(OpCodes.Newobj, constructorBuilder);
        ilGenerator.Emit(OpCodes.Stloc_0);
        ilGenerator.Emit(OpCodes.Ldloc_0);
        ilGenerator.Emit(OpCodes.Ldarg_0);
        ilGenerator.Emit(OpCodes.Ldfld, targetField);
        ilGenerator.Emit(OpCodes.Stfld, innerTargetField);
        for (short i = 1; i <= innerField.Length; i++) {
          ilGenerator.Emit(OpCodes.Ldloc_0);
          if (i <= 3) {
            switch (i) {
              case 1:
                ilGenerator.Emit(OpCodes.Ldarg_1);
                break;
              case 2:
                ilGenerator.Emit(OpCodes.Ldarg_2);
                break;
              case 3:
                ilGenerator.Emit(OpCodes.Ldarg_3);
                break;
            }
          } else {
            ilGenerator.Emit(OpCodes.Ldarg, i);
          }
          ilGenerator.Emit(OpCodes.Stfld, innerField[i - 1]);
        }
        ilGenerator.Emit(OpCodes.Ldarg_0);
        ilGenerator.Emit(OpCodes.Ldfld, controlField);
        ilGenerator.Emit(OpCodes.Ldloc_0);
        ilGenerator.Emit(OpCodes.Ldftn, runMethodBuilder);
        ilGenerator.Emit(OpCodes.Newobj, controlDelegateConstructorBuilder);
        ilGenerator.Emit(OpCodes.Callvirt, beginInvokeMethodInfo);
        ilGenerator.Emit(OpCodes.Pop);
        ilGenerator.Emit(OpCodes.Ret);
      }

      Type proxyType = typeBuilder.CreateType();
      object proxyInstance = Activator.CreateInstance(proxyType, target, control);

      return (T)proxyInstance;
    }

    public static T createProxy<T>(T target) {

      Type interfaceType = typeof(T);
      ConstructorInfo waitCallbackConstructor
          = typeof(WaitCallback).GetConstructor(new Type[] { typeof(object), typeof(IntPtr) });
      MethodInfo queueUserWorkItemMethodInfo
          = typeof(ThreadPool).GetMethod("QueueUserWorkItem", new Type[] { typeof(WaitCallback) });

      AssemblyBuilder assemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(
          new AssemblyName("WorkerThreadProxyAssembly"),
          AssemblyBuilderAccess.Run);
      ModuleBuilder moduleBuilder = assemblyBuilder.DefineDynamicModule(
          "WorkerThreadProxyAssembly.WorkerThreadProxyModule");
      TypeBuilder typeBuilder = moduleBuilder.DefineType(
          "ThreadProxyFactoryNamespace.0ClassProxy",
          TypeAttributes.Class | TypeAttributes.Public,
          typeof(Object),
          new Type[] { interfaceType });

      FieldBuilder targetField = typeBuilder.DefineField("target", interfaceType, FieldAttributes.Private);
      ConstructorBuilder constructorBuilder = typeBuilder.DefineConstructor(
          MethodAttributes.Public | MethodAttributes.HideBySig
              | MethodAttributes.SpecialName | MethodAttributes.RTSpecialName,
          CallingConventions.HasThis | CallingConventions.ExplicitThis,
          new Type[] { interfaceType });
      ILGenerator ilGenerator = constructorBuilder.GetILGenerator();
      ilGenerator.Emit(OpCodes.Ldarg_0);
      ilGenerator.Emit(OpCodes.Call, typeof(object).GetConstructor(new Type[0]));
      ilGenerator.Emit(OpCodes.Ldarg_0);
      ilGenerator.Emit(OpCodes.Ldarg_1);
      ilGenerator.Emit(OpCodes.Stfld, targetField);
      ilGenerator.Emit(OpCodes.Ret);

      foreach (MethodInfo methodInfo in interfaceType.GetMethods()) {
        ParameterInfo[] parameterInfos = methodInfo.GetParameters();
        Type[] parameters = new Type[parameterInfos.Length];
        for (int i = 0; i < parameterInfos.Length; i++) {
          parameters[i] = parameterInfos[i].ParameterType;
        }

        string callClassName = "ThreadProxyFactoryNamespace.00Class" + methodInfo.Name;
        TypeBuilder callClassTypeBuilder = moduleBuilder.DefineType(
            callClassName, TypeAttributes.Class | TypeAttributes.Public);
        constructorBuilder = callClassTypeBuilder.DefineDefaultConstructor(MethodAttributes.Public);
        MethodBuilder runMethodBuilder = typeBuilder.DefineMethod(
            "Run",
            MethodAttributes.Public | MethodAttributes.HideBySig,
            CallingConventions.HasThis | CallingConventions.ExplicitThis,
            typeof(void),
            new Type[] { typeof(object) });
        ilGenerator = runMethodBuilder.GetILGenerator();
        ilGenerator.Emit(OpCodes.Ldarg_0);
        FieldBuilder innerTargetField
            = callClassTypeBuilder.DefineField("0target", interfaceType, FieldAttributes.Public);
        ilGenerator.Emit(OpCodes.Ldfld, innerTargetField);
        FieldBuilder[] innerField = new FieldBuilder[parameterInfos.Length];
        for (int i = 0; i < parameterInfos.Length; i++) {
          ParameterInfo parameterInfo = parameterInfos[i];
          innerField[i] = callClassTypeBuilder.DefineField(
              parameterInfo.Name, parameterInfo.ParameterType, FieldAttributes.Public);
          ilGenerator.Emit(OpCodes.Ldarg_0);
          ilGenerator.Emit(OpCodes.Ldfld, innerField[i]);
        }
        ilGenerator.Emit(OpCodes.Callvirt, methodInfo);
        ilGenerator.Emit(OpCodes.Ret);

        Type callClassType = callClassTypeBuilder.CreateType();

        MethodBuilder methodBuilder = typeBuilder.DefineMethod(
            methodInfo.Name,
            MethodAttributes.Public | MethodAttributes.HideBySig | MethodAttributes.NewSlot
                | MethodAttributes.Virtual | MethodAttributes.Final,
            CallingConventions.HasThis | CallingConventions.ExplicitThis,
            methodInfo.ReturnType,
            parameters);
        ilGenerator = methodBuilder.GetILGenerator();
        ilGenerator.DeclareLocal(callClassType);
        ilGenerator.Emit(OpCodes.Newobj, constructorBuilder);
        ilGenerator.Emit(OpCodes.Stloc_0);
        ilGenerator.Emit(OpCodes.Ldloc_0);
        ilGenerator.Emit(OpCodes.Ldarg_0);
        ilGenerator.Emit(OpCodes.Ldfld, targetField);
        ilGenerator.Emit(OpCodes.Stfld, innerTargetField);
        for (short i = 1; i <= innerField.Length; i++) {
          ilGenerator.Emit(OpCodes.Ldloc_0);
          if (i <= 3) {
            switch (i) {
              case 1:
                ilGenerator.Emit(OpCodes.Ldarg_1);
                break;
              case 2:
                ilGenerator.Emit(OpCodes.Ldarg_2);
                break;
              case 3:
                ilGenerator.Emit(OpCodes.Ldarg_3);
                break;
            }
          } else {
            ilGenerator.Emit(OpCodes.Ldarg, i);
          }
          ilGenerator.Emit(OpCodes.Stfld, innerField[i - 1]);
        }
        ilGenerator.Emit(OpCodes.Ldloc_0);
        ilGenerator.Emit(OpCodes.Ldftn, runMethodBuilder);
        ilGenerator.Emit(OpCodes.Newobj, waitCallbackConstructor);
        ilGenerator.Emit(OpCodes.Call, queueUserWorkItemMethodInfo);
        ilGenerator.Emit(OpCodes.Pop);
        ilGenerator.Emit(OpCodes.Ret);
      }

      Type proxyType = typeBuilder.CreateType();
      object proxyInstance = Activator.CreateInstance(proxyType, target);

      return (T)proxyInstance;
    }
  }
}
